#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 15 09:59:25 2022

@author: nurmelaj
"""

import time
from datetime import datetime
import xarray as xr
import numpy as np
import pandas as pd
from scipy.optimize import least_squares
import warnings
warnings.filterwarnings('ignore', category=RuntimeWarning)

import ddeq


def estimate_emissions(datasets, wind_folder, sources,
                      lon_km=50, lat_km=50, grid_reso=5,
                      wind_product='ERA5',
                      start_date='2015-01-01', end_date='2015-12-24',
                      AM_options={'warmup':10000,'samples':90000},
                      trace_gases=['CO2']):
    """ 
    Estimates the NOx and CO2 emissions for one source over the chosen time series
    by computing the divergence map using the SMARTCARB data with the chosen satellites
    and fitting a peak fitting model.
    """
    # Process start time
    estimation_start = time.process_time()

    # Dates
    start_date = pd.Timestamp(start_date)
    end_date = pd.Timestamp(end_date)
    dates = pd.date_range(start_date, end_date)

    # Validate supported trace gases
    for gas in trace_gases:
        if gas not in ['CO2','NO2']:
            raise ValueError(f'Trace gas {gas} not supported')

    if wind_product not in ['ERA5','SMARTCARB']:
        raise ValueError(f'Unknown wind product {wind_product}')

    results_ds = ddeq.misc.init_results_dataset(sources, trace_gases,
                                                units='Mt/y',
                                                method='divergence')

    for source_name, source in sources.groupby('source'):

        for gas in trace_gases:
            print(f"Computing divergence map for source '{source_name}'...")
            start_time = time.process_time()

            # First date, ignore timestamp
            date = dates[0].date()
            center_lon, center_lat = float(source['lon_o']), float(source['lat_o'])

            # Regular km grid around the source
            longrid, latgrid, longrid_km, latgrid_km = ddeq.misc.grids(center_lon,center_lat,lon_km,lat_km,grid_reso)

            # Area limits
            latlims, lonlims = (np.min(latgrid), np.max(latgrid)), (np.min(longrid), np.max(longrid))

            # Gridded eastward and northward fluxes
            flux_xr = []

            for date in dates:
                date = date.date()
                wind = ddeq.wind.read_gnfra_wind(
                    wind_folder, date, latlims=latlims, lonlims=lonlims,
                    dataset=wind_product
                )
                flux_xr_next = ddeq.divergence_model.gridded_flux(
                    datasets, wind, date,
                    latgrid, longrid, latgrid_km, longrid_km,
                    wind_product=wind_product,
                    species=gas
                )
                flux_xr.append(flux_xr_next)

            flux_xr = xr.concat(flux_xr, 'n_images')

            # Compute divergence map from the temporally averaged fluxes
            div_xr = ddeq.divergence_model.compute_divergence_map(flux_xr,species=gas)
            end_time = time.process_time()
            print(f'...done in {end_time-start_time} seconds')

            print(f"Optimizing trace gas {gas} emissions for source '{source_name}'...")
            start_time = time.process_time()
            fitting_xr = peak_fitting(div_xr,species=gas,AM_options=AM_options)
            end_time = time.process_time()
            print(f'...done in {end_time-start_time} seconds')

            # Emission estimation and standard deviation in g/s
            emis_est, emis_std = fitting_xr['opt'].data[0], np.sqrt(fitting_xr['cov'].data[0,0])

            # Conversion to Mt/y
            emis_est, emis_std = emis_est*3600*24*365.24219/10**12, emis_std*3600*24*365.24219/10**12
            # Write results to xarray dataset
            results_ds[f'{gas}_estimated_emissions'].loc[dict(source=source_name)] = emis_est
            results_ds[f'{gas}_estimated_emissions_std'].loc[dict(source=source_name)] = emis_std

    estimation_end = time.process_time()      
    print(f'Estimations done in {estimation_end-estimation_start} seconds')
    return results_ds


def peak_fitting(div_xr,species='CO2',pixels=4,AM_options=None):

    latgrid_km, longrid_km = div_xr['latgrid_km'].data, div_xr['longrid_km'].data

    # Create pixels x pixels sized kilometer grid around the source
    row_slice = slice((len(latgrid_km)-1)//2-pixels,(len(latgrid_km)+1)//2+pixels,1)
    col_slice = slice((len(longrid_km)-1)//2-pixels,(len(longrid_km)+1)//2+pixels,1)

    # Corresponding 2d X and Y kilometer grids around the source
    Y, X = np.meshgrid(latgrid_km[row_slice],longrid_km[col_slice],indexing='ij')

    # Intial values
    if species == 'NO2':
        init = [900, 5, 6, 0.0, 0.0, 0.0, -0.5]
        bounds = np.array([(0,2000), (1,9), (1,9), (-10,10), (-10,10), (-1,1), (-1,1)])
        scales = [1e3,1e0,1e0,1e0,1e0,1e0,1e-1]
    elif species == 'CO2':
        init = [9e5, 7, 7, 0.0, 0.0, 0.0, -500]
        bounds = np.array([(0,2e6), (1,9), (1,9), (-10,10), (-10,10), (-1,1), (-1000,1000)])
        scales = [1e5,1e0,1e0,1e0,1e0,1e0,1e2]

    # Divergence near the source
    div_map = div_xr[f'{species}_div'].data[row_slice, col_slice]

    # Residual vector or component-wise difference between the model and the data
    residual = lambda params: ddeq.divergence_model.model_residual(params,(X,Y),div_map)

    # Least square optimization using scipy with default Trust Region Reflective algorithm
    optimized = least_squares(residual, init, x_scale=scales, bounds=bounds.T, method='trf')
    opt = optimized.x
    J = optimized.jac
    cov = np.linalg.inv(J.T @ J)

    # Optimization using adaptive metropolis algorithm if AM_options are given
    if AM_options is not None:
        warmup, samples = AM_options['warmup'], AM_options['samples']
        sampling = ddeq.mcmc_tools.AM_MC(residual,opt,samples,warmup,C0=cov,
                         bounds=bounds,init_fit=False,progress=False)
        opt, cov = sampling['MAP'].data, sampling['cov'].data

    names = ['peak_integral','sigma_x','sigma_y','x0','y0','corr','background']
    ds = xr.Dataset({'opt': xr.DataArray(opt,dims=['params'],coords=dict(params=names)),
                     'cov': xr.DataArray(cov,dims=['params','params'],coords=dict(params=names)),
                     'bg': opt[-1]})
    return ds