#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 15 10:48:43 2022

@author: nurmelaj
"""

import numpy as np
import xarray as xr
import glob
import fnmatch

import ddeq
from ddeq.misc import earth
from ddeq.wind import spatial_interpolation


def gridded_flux(datasets, wind, date, latgrid, longrid, latgrid_km, longrid_km,
                 wind_product='ERA5', species='CO2'):

    # Shape of the area
    shape = (len(latgrid_km),len(longrid_km))

    # Initiate gridded array for chosen species
    species_gridded = np.full(shape,np.nan)
    species_Fx_flux = np.full(shape,np.nan)
    species_Fy_flux = np.full(shape,np.nan)

    images = []
    species_grids = []
    flux_Fx = []
    flux_Fy = []

    for i, file in enumerate(datasets.read_date(date)):

        filename = datasets.get_filenames(date)[i]

        psurf, lon, lat = file.psurf.data.flatten(), file.lon.data.flatten(), file.lat.data.flatten()

        if species == 'CO2':
            data = file.CO2.data

            # Convolution for CO2
            data = ddeq.misc.normalized_convolution(data, np.ones(shape=(5,5))/25).flatten()

        elif species == 'NO2':
            data = file.NO2.data.flatten()
            # NO2 units from molec/cm^2 to molec/m^2
            data = 10000 * data

        inside_ind = indices_inside_grid(lon,lat,longrid,latgrid,enlarge=0.1)

        if len(inside_ind) == 0:
            continue

        psurf, lon, lat = psurf[inside_ind], lon[inside_ind], lat[inside_ind]
        data = data[inside_ind]

        # Check if the area contains any data for the chosen species
        if np.all(np.isnan(data)):
            continue

        # Append the file name/image
        images.append(filename)

        # XCO2 anomaly inside the chosen grid, excluding points with distance less than 10 km to the grid center
        center_lon = longrid[(shape[0]-1)//2,(shape[1]-1)//2]
        center_lat = latgrid[(shape[0]-1)//2,(shape[1]-1)//2]

        # Distances in metres of each data pixel to the center point of the grid
        dists = np.asarray(earth.inverse(
            np.repeat([(center_lon,center_lat)], len(data),axis=0),
            np.column_stack((lon,lat))))[:,0]

        # Remove CO2 background
        if species == 'CO2':
            data = data - np.nanmedian(data[np.where(dists >= 10000)])

            # Change CO2 units from ppm to molec/m^2
            data = data*6.0221409e23/9.807/0.02896*psurf/1e6

        elif species == 'NO2':
            # Change NO2 units from molec/cm^2 to molec/m^2
            data = 10000 * data

        # Get wind data from xarray dataset
        wu = wind['U_GNFR_A'][0,:,:].data
        wv = wind['V_GNFR_A'][0,:,:].data

        if wind_product == 'ERA5':
            wlon = wind['longitude'][:].data
            wlat = wind['latitude'][:].data
        elif wind_product == 'SMARTCARB':
            wlon = wind['lon'][:,:].data
            wlat = wind['lat'][:,:].data

        # Interpolate wind spatially in spherical coordinates
        wu_interp = spatial_interpolation(wu,wlon,wlat,lon,lat)
        wv_interp = spatial_interpolation(wv,wlon,wlat,lon,lat)

        # Fluxes at each data coordinate
        Fx = data * wu_interp
        Fy = data * wv_interp

        # Gridding using undersampling, grid resolution should be about the same
        # or less than the data resolution, each value of the grid is an average
        # of the values inside the grid pixel
        for r in range(shape[0]):
            if r == 0:
                lat_diff_north = np.mean(latgrid[0,:] - latgrid[1,:])
                lat_diff_south = latgrid[r,0]-latgrid[r+1,0]
            elif r == shape[0]-1:
                lat_diff_north = latgrid[r-1,0]-latgrid[r,0]
                lat_diff_south = np.mean(latgrid[-2,:] - latgrid[-1,:])
            else:
                lat_diff_north = latgrid[r-1,0]-latgrid[r,0]
                lat_diff_south = latgrid[r,0]-latgrid[r+1,0]
            ind_lat = np.where((latgrid[r,0]-lat_diff_south/2 <= lat) & (lat <= latgrid[r,0]+lat_diff_north/2))[0]
            if not ind_lat.any():
                continue
            for c in range(shape[1]):
                if c == 0:
                    lon_diff_east = longrid[r,c+1]-longrid[r,c]
                    lon_diff_west = longrid[r,1] - longrid[r,0]
                elif c == shape[1]-1:
                    lon_diff_east = longrid[r,-1]-longrid[r,-2]
                    lon_diff_west = longrid[r,c]-longrid[r,c-1]
                else:
                    lon_diff_east = longrid[r,c+1]-longrid[r,c]
                    lon_diff_west = longrid[r,c]-longrid[r,c-1]
                ind_lon = np.where((longrid[r,c]-lon_diff_west/2 <= lon[ind_lat]) & (lon[ind_lat] <= (longrid[r,c]+lon_diff_east/2)))[0]
                if not ind_lon.any():
                    continue
                # Average data values inside each grid pixels
                species_gridded[r,c] = np.nanmean(data[ind_lat][ind_lon]) if np.isnan(species_gridded[r,c]) else species_gridded[r,c] + np.nanmean(data[ind_lat][ind_lon])
                species_Fx_flux[r,c] = np.nanmean(Fx[ind_lat][ind_lon]) if np.isnan(species_Fx_flux[r,c]) else species_Fx_flux[r,c] + np.nanmean(Fx[ind_lat][ind_lon])
                species_Fy_flux[r,c] = np.nanmean(Fy[ind_lat][ind_lon]) if np.isnan(species_Fy_flux[r,c]) else species_Fy_flux[r,c] + np.nanmean(Fy[ind_lat][ind_lon])
        species_grids.append(species_gridded)
        flux_Fx.append(species_Fx_flux)
        flux_Fy.append(species_Fy_flux)

    images_xr = xr.DataArray(images,dims=['n_images'])
    if images_xr.size != 0:
        grid_xr = xr.DataArray(species_grids,dims=['n_images','lat','lon'],attrs={'long name':f'gridded {species}','units':'molec/m2'})
        Fx_xr = xr.DataArray(flux_Fx,dims=['n_images','lat','lon'],attrs={'long name':f'Eastward flux of {species}','units':'molec*m/s'})
        Fy_xr = xr.DataArray(flux_Fy,dims=['n_images','lat','lon'],attrs={'long name':f'Northward flux of {species}','units':'molec*m/s'})
    else:
        grid_xr = xr.DataArray(np.full((0,shape[0],shape[1]),np.nan),dims=['n_images','lat','lon'],attrs={'long name':f'gridded {species}','units':'molec/m2'})
        Fx_xr = xr.DataArray(np.full((0,shape[0],shape[1]),np.nan),dims=['n_images','lat','lon'],attrs={'long name':f'Eastward flux of {species}','units':'molec*m/s'})
        Fy_xr = xr.DataArray(np.full((0,shape[0],shape[1]),np.nan),dims=['n_images','lat','lon'],attrs={'long name':f'Northward flux of {species}','units':'molec*m/s'})
    # Save to xarray dataset
    data_vars = {f'{species}_grid' : grid_xr,
                 f'{species}_Fx' : Fx_xr,
                 f'{species}_Fy' : Fy_xr,
                 'longrid': xr.DataArray(longrid,dims=['lat','lon']),
                 'latgrid': xr.DataArray(latgrid,dims=['lat','lon']),
                 'longrid_km': xr.DataArray(longrid_km,dims=['lon']),
                 'latgrid_km': xr.DataArray(latgrid_km,dims=['lat']),
                 'images': images_xr}
    attrs = {'Description': f'Gridded {species} data values and fluxes'}
    return xr.Dataset(data_vars=data_vars,attrs=attrs)


def compute_divergence_map(flux_xr,species='CO2'):
    '''
    Computes divergence from fluxes averaged over different data files.

    Parameters
    ----------
    flux_xr : xarray dataset
        Precomputed species fluxes.
    species : str, optional
        Chosen species, either 'CO2' or 'NO2'. The default is 'CO2'.

    Returns
    -------
    xarray-dataset
        Computed divergence for the chosen species.
    '''
    Fx = flux_xr[f'{species}_Fx'].mean(dim='n_images',skipna=True).data
    Fy = flux_xr[f'{species}_Fy'].mean(dim='n_images',skipna=True).data
    longrid_km = flux_xr['longrid_km'][dict(n_images=0)]
    latgrid_km = flux_xr['latgrid_km'][dict(n_images=0)]
    if species == 'CO2':
        # Divergence of CO2 in g/m²/s or Mg/km²/s
        div = 44.01/6.0221409e23*divergence([Fx,Fy],[longrid_km.data*1e3,latgrid_km.data*1e3])
    elif species == 'NO2':
        # NO2 fluxes with NOx correction, about 95.71 % of NOx is NO2
        Fx = 0.9571*Fx
        Fy = 0.9571*Fy
        # Divergence of NO2 in g/m²/s or Mg/km²/s
        div = 46.0055/6.0221409e23*divergence([Fx,Fy], [longrid_km.data*1e3,latgrid_km.data*1e3])
        '''
        # NO2 Sinks in g/m²/s assuming 4h lifetime
        sinks = 46.0055/6.0221409e23*0.9571*species_gridded/(4*3600)
        sinks = sinks*1e6
        # Corresponding NO2 emissions
        emissions = div+sinks
        '''
    # Conversion from Mg/km²/s to g/km²/s
    div = 1e6*div
    # Return as xarray dataset
    div_xr = xr.DataArray(div,dims=['lat','lon'],attrs={'long name':f'Gridded {species} divergence','units':'g/m2/s'})
    data_vars = {'longitude': flux_xr['longrid'][dict(n_images=0)],
                 'latitude': flux_xr['latgrid'][dict(n_images=0)],
                 'longrid_km': longrid_km,
                 'latgrid_km': latgrid_km,
                 f'{species}_div': div_xr,
                 f'{species}_grid': flux_xr[f'{species}_grid'].mean(dim='n_images',skipna=True)}
    attrs = {'Description': f'Gridded {species} divergence'}
    return xr.Dataset(data_vars=data_vars,attrs=attrs)
    
def divergence(f, sp):
    """ 
    Computes divergence of vector field 
    f: array -> vector field components [Fx,Fy,Fz,...]
    sp: array -> spacing between points in respective directions 
    [spx,spy,spz,...]
    """
    d = len(f)
    grad = [np.gradient(f[i], sp[i], axis=d-1-i) for i in range(d)]
    return np.ufunc.reduce(np.add, grad)
    
def indices_inside_grid(data_x,data_y,grid_x,grid_y,enlarge=0.0):
    if grid_x.ndim == 1 and grid_y.ndim == 1:
        indices = np.where((np.min(grid_x)-enlarge <= data_x) & (data_x <= np.max(grid_x)+enlarge) &
                           (np.min(grid_y)-enlarge <= data_y) & (data_y <= np.max(grid_y)+enlarge))[0]
    elif grid_x.ndim == 2 and grid_y.ndim == 2:
        # Four lines limiting the grid
        top_line = lambda x : (grid_y[0,-1]-grid_y[0,0]) / (grid_x[0,-1]-grid_x[0,0]) * (x-grid_x[0,0]) + grid_y[0,0] + enlarge
        bottom_line = lambda x : (grid_y[-1,-1]-grid_y[-1,0]) / (grid_x[-1,-1]-grid_x[-1,0]) * (x-grid_x[-1,-1]) + grid_y[-1,-1] - enlarge
        left_line = lambda y : (grid_x[-1,0]-grid_x[0,0]) / (grid_y[-1,0]-grid_y[0,0]) * (y-grid_y[0,0]) + grid_x[0,0] - enlarge
        right_line = lambda y : (grid_x[-1,-1]-grid_x[0,-1]) / (grid_y[-1,-1]-grid_y[0,-1]) * (y-grid_y[0,-1]) + grid_x[0,-1] + enlarge
        indices = np.where((bottom_line(data_x) <= data_y) & (data_y <= top_line(data_x)) & 
                           (left_line(data_y) <= data_x) & (data_x <= right_line(data_y)))[0]
    return indices

def peak_model(theta, grids):
    """
    Model that we want to fit to each source
    
    theta = parameters that to be optimized
    grid = cropped grid around the source
    """
    # Peak integral corresponding emission in kg/s
    Q = theta[0]
    # Deviations x and y directions in km
    sigma_x, sigma_y = theta[1], theta[2]
    # Center of the peak
    x0, y0 = theta[3], theta[4]
    # Correlation between x and y dimensions
    corr = theta[5]
    # Background in kg/m²/s
    B = theta[6]
    # Kilometer grids
    X, Y = grids[0], grids[1]
    
    #mx = theta[7]
    #my = theta[8]
    #mx, my = 0, 0
    
    #G = A/(2*np.pi*sigma_x*sigma_y)*np.exp(-((X-x0)**2/(2*sigma_x**2)+(Y-y0)**2/(2*sigma_y**2)))+mx*(X-x0)+my*(Y-y0)+B
    # Normalization
    N = Q/(2*np.pi*sigma_x*sigma_y*np.sqrt(1-corr**2))
    G = N*np.exp(-((X-x0)**2/(2*sigma_x**2*(1-corr**2))+(Y-y0)**2/(2*sigma_y**2*(1-corr**2))-corr*(X-x0)*(Y-y0)/(sigma_x*sigma_y*(1-corr**2)))) + B
    #G += mx*(X-x0)+my*(Y-y0)+B
    return G

def model_residual(theta, grids, data):
    """ 
    Residual function.
    """
    G = peak_model(theta, grids)
    non_nan = ~np.isnan(data)
    data = data[non_nan]
    residual = (G[non_nan] - data).ravel()
    #sigma_obs = theta[-1]
    #ss = 0.5*np.sum(np.square(residual))/sigma_obs**2 + 0.5*len(residual)*np.log(2*np.pi) + len(residual)*np.log(sigma_obs)
    return residual